(** Combinator Parsers *)
(* type t *)
type ('a,'t)t = 't list -> ('a * 't list)option
(* *)

(* Primitive Parsers *)
(* eoi : (unit,'t)t *)
let eoi = function [] -> Some ((),[])
                 | _  -> None
(* return : 'a -> ('a,'t)t *)
let return x = fun ts -> Some (x,ts)
(* token : ('t -> 'a option) -> ('a,'t)t *)
let token f = function
  | []    -> None
  | t::ts -> match f t with
    | Some x -> Some (x,ts)
    | None   -> None
(* tokens : 't list -> ('t list,'t)t *)
let tokens ts input =
  let rec tokens = function ([],ss)                  -> Some (ts,ss)
                          | (t::ts,s::ss) when s = t -> tokens (ts,ss)
                          | _                        -> None
  in
  tokens(ts,input)
(* *)

(* Parsers *)
(* sat : ('t -> bool) -> ('a,'t)t *)
let sat p ts = token(fun t -> if p t then Some t else None) ts
(* *)

(* Primitive Combinators *)
(* bind : ('a,'t)t -> ('a -> ('b,'t)t) -> ('b,'t)t *)
let (>>=) p f ts = match p ts with None      -> None
                                | Some(x,ts) -> f x ts
(* then : ('a,'t)t -> ('b,'t)t -> ('b,'t)t *)
let (>>) p q ts = (>>=) p (fun _ -> q) ts
(* choice : ('a,'t)t -> ('a,'t)t -> ('a,'t)t *)
let (<|>) p q ts = match p ts with None        -> q ts
                                 | Some _ as r -> r
(* *)

(* Combinators *)
(* optional : ('a,'t)t -> 'a -> ('a,'t)t *)
let (?>) p d = p <|> return d
(* between : ('a,'t)t -> ('b,'t)t -> ('c,'t)t -> ('b,'t)t *)
let between l p r =
  l >>= fun _ ->
  p >>= fun x ->
  r >>= fun _ ->
  return x
(* many : ('a,'t)t -> ('a list,'t)t *)
let rec many p = (
  p      >>= fun x ->
  many p >>= fun xs ->
  return (x::xs)
) <|> return []
(* many1 : ('a,'t)t -> ('a list,'t)t *)
let many1 p =
  p      >>= fun x ->
  many p >>= fun xs ->
  return (x::xs)
(* sep_by1 : ('a,'t)t -> ('b,'t)t -> ('b list,'t)t *)
let sep_by1 sep p =
  p                        >>= fun x ->
  many (sep >>= fun _ -> p) >>= fun xs ->
  return (x::xs)
(* sep_by : ('a,'t)t -> ('b,'t)t -> ('b list,'t)t *)
let sep_by sep p = sep_by1 sep p <|> return []
(* *)

(* Character Parsers *)
(* any : (char,char)t *)
let any ts = sat (fun _ -> true) ts
(* char : char -> (char,char)t *)
let char c ts = sat ((=) c) ts
(* digit : (char,char)t *)
let digit = sat (fun c -> '0' <= c && c <= '9')
(* letter : (char,char)t *)
let letter = 
 sat(fun c -> ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z'))
(* noneof : string -> (char,char)t *)
let noneof s = sat (fun c -> not(Lst.mem c (Strng.of_string s)))
(* oneof : string -> (char,char)t *)
let oneof s = sat (fun c -> Lst.mem c (Strng.of_string s))
(* space : (char,char)t *)
let space = oneof " \n\r\t"
(* spaces : (int,char)t *)
let spaces = many space >>= fun ls -> return (Lst.length ls)
(* string : string -> (char list,char)t *)
let string s = tokens (Strng.of_string s)
(* *)

(** Running Parsers on Input *)
(* test : ('a,'t)t -> 't list -> 'a *)
let test p s = match p (Strng.of_string s) with
  | None      -> failwith "parse error"
  | Some (x,ts) -> (x,ts)
(* parse : ('a,'t)t -> 't list -> 'a option *)
let parse p input = match p input with None      -> None
                                     | Some (x,_) -> Some x
